import aiofiles
import os
import re
import shutil
from typing import List


async def _load_sysctl_cpus(hub) -> int:
    sysctl = shutil.which("sysctl")
    if sysctl:
        ret = await hub.exec.cmd.run([sysctl, "-n", "hw.ncpu"])
        if not ret.retcode:
            return int(ret.stdout)


async def _load_sysctl_arch(hub) -> str:
    sysctl = shutil.which("sysctl")
    if sysctl:
        ret = await hub.exec.cmd.run([sysctl, "-n", "hw.machine"])
        if not ret.retcode:
            return ret.stdout.strip()


async def _load_sysctl_model(hub) -> str:
    sysctl = shutil.which("sysctl")
    if sysctl:
        ret = await hub.exec.cmd.run([sysctl, "-n", "hw.model"])
        if not ret.retcode:
            return ret["stdout"].strip()


async def _load_sysctl_flags(hub) -> str:
    sysctl = shutil.which("sysctl")
    if sysctl:
        ret = await hub.exec.cmd.run([sysctl, "-n", "machdep.cpu.features"])
        if not ret.retcode:
            return ret.stdout.strip().split(" ")


async def _load_arch(hub) -> str:
    arch = shutil.which("arch")
    if arch:
        ret = await hub.exec.cmd.run([arch, "-s"])
        if not ret.retcode:
            return ret.stdout.strip()


async def _load_cpuctl_flags(hub) -> List[str]:
    cpu_flags = []
    cputil = shutil.which("cputil")
    if cputil:
        ret = await hub.exec.cmd.run([cputil, "identify", "0"])
        for line in ret.stdout.splitlines():
            cpu_match = re.match(r"cpu[0-9]: features[0-9]? .+<(.+)>", line)
            if cpu_match:
                flag = cpu_match.group(1).split(",")
                cpu_flags.extend(flag)

    return cpu_flags


async def _load_dmesg_boot_flags(hub) -> List[str]:
    cpu_flags = []
    dmesg_boot = "/var/run/dmesg.boot"

    if os.path.exists(dmesg_boot):
        async with aiofiles.open(dmesg_boot, "r") as _fp:
            cpu_here = False
            async for line in _fp:
                if line.startswith("CPU: "):
                    cpu_here = True  # starts CPU descr
                    continue
                if cpu_here:
                    if not line.startswith(" "):
                        break  # game over
                    if "Features" in line:
                        start = line.find("<")
                        end = line.find(">")
                        if start > 0 and end > 0:
                            flag = line[start + 1 : end].split(",")
                            cpu_flags.extend(flag)
    return cpu_flags


async def load_cpu_arch(hub):
    hub.corn.CORN.cpuarch = (
        await _load_sysctl_arch(hub)
        or await _load_arch(hub)
        or hub.corn.CORN.get("osarch")
    )


async def load_cpu_flags(hub):
    hub.corn.CORN.cpu_flags = sorted(
        map(
            str.lower,
            set(await _load_cpuctl_flags(hub) or await _load_dmesg_boot_flags(hub)),
        )
    )

    # Report if hardware virtualization is available under amd or intel
    hub.corn.CORN.hardware_virtualization = any(
        f in hub.corn.CORN.cpu_flags for f in ("svm", "vmx")
    )


async def load_cpu_model(hub):
    hub.corn.CORN.cpu_model = await _load_sysctl_model(hub)


async def load_num_cpus(hub):
    hub.corn.CORN.num_cpus = await _load_sysctl_cpus(hub) or 1



"""
Get system specific hardware data from dmidecode

Provides
    biosversion
    productname
    manufacturer
    serialnumber
    biosreleasedate
    uuid

.. versionadded:: 0.9.5
"""
import shutil
from typing import Dict


async def _load_kenv(hub) -> Dict[str, str]:
    # For FreeBSD
    kenv = shutil.which("kenv")
    if kenv:
        return {
            "biosversion": (
                await hub.exec.cmd.run([kenv, "smbios.bios.version"])
            ).stdout.strip(),
            "manufacturer": (
                await hub.exec.cmd.run([kenv, "smbios.system.maker"])
            ).stdout.strip(),
            "serialnumber": (
                await hub.exec.cmd.run([kenv, "smbios.system.serial"])
            ).stdout.strip(),
            "productname": (
                await hub.exec.cmd.run([kenv, "smbios.system.product"])
            ).stdout.strip(),
            "biosreleasedate": (
                await hub.exec.cmd.run([kenv, "smbios.bios.reldate"])
            ).stdout.strip(),
            "uuid": (
                await hub.exec.cmd.run([kenv, "smbios.system.uuid"])
            ).stdout.strip(),
        }


async def _load_sysctl(hub) -> Dict[str, str]:
    # For OpenBSD and NetBSD
    sysctl = shutil.which("sysctl")
    if sysctl:
        return {
            "biosversion": (
                await hub.exec.cmd.run([sysctl, "hw.version"])
            ).stdout.strip()
            or (
                await hub.exec.cmd.run([sysctl, "machdep.dmi.board-version"])
            ).stdout.strip(),
            "manufacturer": (
                await hub.exec.cmd.run([sysctl, "hw.vendor"])
            ).stdout.strip()
            or (
                await hub.exec.cmd.run([sysctl, "machdep.dmi.system-vendor"])
            ).stdout.strip(),
            "serialnumber": (
                await hub.exec.cmd.run([sysctl, "hw.serialno"])
            ).stdout.strip()
            or (
                await hub.exec.cmd.run([sysctl, "machdep.dmi.system-serial"])
            ).stdout.strip(),
            "productname": (
                await hub.exec.cmd.run([sysctl, "hw.product"])
            ).stdout.strip()
            or (
                await hub.exec.cmd.run([sysctl, "machdep.dmi.system-product"])
            ).stdout.strip(),
            "biosreleasedate": (
                await hub.exec.cmd.run([sysctl, "machdep.dmi.bios-date"])
            ).stdout.strip(),
            "uuid": (await hub.exec.cmd.run([sysctl, "hw.uuid"])).stdout.strip()
            or (
                await hub.exec.cmd.run([sysctl, "machdem.dmi.system-uuid"])
            ).stdout.strip(),
        }


async def load_bios_data(hub):
    data = await _load_kenv(hub) or await _load_sysctl(hub) or ("unknown" * 6)

    for key, val in data.items():
        hub.corn.CORN[key] = await hub.corn.init.clean_value(key, val)



import shutil


async def _load_swapctl(hub) -> int:
    swapctl = shutil.which("swapctl")

    if swapctl:
        swap_data = (await hub.exec.cmd.run([swapctl, "-sk"])).stdout.strip()
        if swap_data != "no swap devices configured":
            return int(swap_data.split()[1]) // 1024


async def _load_sysctl_swap(hub) -> int:
    sysctl = shutil.which("sysctl")
    if sysctl:
        return (
            int((await hub.exec.cmd.run([sysctl, "-n", "vm.swap_total"])).stdout)
            // 1024
            // 1024
        )


async def load_swap(hub):
    """
    Return the swap information for BSD-like systems
    """
    hub.corn.CORN.swap_total = (
        await _load_swapctl(hub) or await _load_sysctl_swap(hub) or 0
    )


async def load_meminfo(hub):
    """
    Return the memory information for BSD-like systems
    """
    hub.corn.CORN.mem_total = 0

    sysctl = shutil.which("sysctl")

    if sysctl:
        mem = int((await hub.exec.cmd.run([sysctl, "-n", "hw.physmem"])).stdout)
        if mem < 0:
            mem = int((await hub.exec.cmd.run([sysctl, "-n", "hw.physmem64"])).stdout)
        hub.corn.CORN.mem_total = mem // 1024 // 1024



import aiofiles
import logging
import os
import shutil

log = logging.getLogger(__name__)


async def load_init(hub):
    log.debug("Adding init grain")
    hub.corn.CORN.init = "init"

    system = "/run/systemd/system"
    cmdline = "/proc/1/cmdline"

    if shutil.which("rc"):
        # The first choice for most BSD systems
        hub.corn.CORN.init = "rc"
    elif os.path.exists(system) and os.stat(system):
        hub.corn.CORN.init = "systemd"
    elif os.path.exists(cmdline):
        async with aiofiles.open("/proc/1/cmdline") as fhr:
            init_cmdline = (await fhr.read()).replace("\x00", " ").split()
            if len(init_cmdline) >= 1:
                init_bin = shutil.which(init_cmdline[0])
                if init_bin.endswith("bin/init"):
                    supported_inits = (b"upstart", b"sysvinit", b"systemd")
                    edge_len = max(len(x) for x in supported_inits) - 1
                    buf_size = hub.OPT.get("file_buffer_size", 262144)
                    try:
                        async with aiofiles.open(init_bin, "rb") as fp_:
                            edge = b""
                            buf = (await fp_.read(buf_size)).lower()
                            while buf:
                                if isinstance(buf, str):
                                    # This makes testing easier
                                    buf = buf.encode()
                                buf = edge + buf
                                for item in supported_inits:
                                    if item in buf:
                                        item = item.decode("utf-8")
                                        hub.corn.CORN.init = item
                                        buf = b""
                                        break
                                edge = buf[-edge_len:]
                                buf = (await fp_.read(buf_size)).lower()
                    except (IOError, OSError) as exc:
                        log.error(
                            "Unable to read from init_bin (%s): %s", init_bin, exc
                        )
                elif shutil.which("supervisord") in init_cmdline:
                    hub.corn.CORN.init = "supervisord"
                elif shutil.which("dumb-init") in init_cmdline:
                    # https://github.com/Yelp/dumb-init
                    hub.corn.CORN.init = "dumb-init"
                elif shutil.which("tini") in init_cmdline:
                    # https://github.com/krallin/tini
                    hub.corn.CORN.init = "tini"
                elif "runit" in init_cmdline:
                    hub.corn.CORN.init = "runit"
                elif "/sbin/my_init" in init_cmdline:
                    # Phusion Base docker container use runit for srv mgmt, but
                    # my_init as pid1
                    hub.corn.CORN.init = "runit"
                else:
                    log.debug(
                        "Could not determine init system from command line: (%s)",
                        " ".join(init_cmdline),
                    )
            else:
                # Emtpy init_cmdline
                log.warning("Unable to fetch data from /proc/1/cmdline")



import os


async def load_uname(hub):
    """
    Verify that POP-BSD is running on BSD
    """
    a = 1
    (
        hub.corn.CORN.kernel,
        hub.corn.CORN.nodename,
        hub.corn.CORN.kernelrelease,
        hub.corn.CORN.kernelversion,
        hub.corn.CORN.osarch,
    ) = os.uname()

    assert hub.corn.CORN.kernel.upper().endswith(
        "BSD"
    ), "idem-bsd is only intended for BSD systems"

    # Hard coded grains for BSD
    hub.corn.CORN.os_family = "BSD"
    hub.corn.CORN.ps = "ps auxwww"



# -*- coding: utf-8 -*-
from typing import Any, Dict, List


async def sig_run(
    hub,
    cmd: str or List[str],
    cwd: str,
    stdin: str,
    stdout: int,
    stderr: int,
    shell: bool,
    env: Dict[str, Any],
    umask: str,
    timeout: int or float,
    **kwargs,
) -> Dict[str, Any]:
    pass



import logging
import re
import shutil
from typing import Any, Dict, List, Tuple

log = logging.getLogger(__name__)


class GEOM:
    GEOMNAME = "Geom name"
    MEDIASIZE = "Mediasize"
    SECTORSIZE = "Sectorsize"
    STRIPESIZE = "Stripesize"
    STRIPEOFFSET = "Stripeoffset"
    DESCR = "descr"  # model
    LUNID = "lunid"
    LUNNAME = "lunname"
    IDENT = "ident"  # serial
    ROTATIONRATE = "rotationrate"  # RPM or 0 for non-rotating

    # Preserve the API where possible with Salt < 2016.3
    _aliases = {
        DESCR: "device_model",
        IDENT: "serial_number",
        ROTATIONRATE: "media_RPM",
        LUNID: "WWN",
    }

    _datatypes = {
        MEDIASIZE: ("re_int", r"(\d+)"),
        SECTORSIZE: "try_int",
        STRIPESIZE: "try_int",
        STRIPEOFFSET: "try_int",
        ROTATIONRATE: "try_int",
    }


_geom_attribs = [GEOM.__dict__[key] for key in GEOM.__dict__ if not key.startswith("_")]


def _datavalue(datatype, data):
    if datatype == "try_int":
        try:
            return int(data)
        except ValueError:
            return None
    elif datatype is tuple and datatype[0] == "re_int":
        search = re.search(datatype[1], data)
        if search:
            try:
                return int(search.group(1))
            except ValueError:
                return None
        return None
    else:
        return data


async def _parse_geom_attribs(device: str) -> Tuple[Dict[str, Any], List[str]]:
    disks = {}
    ssds = []
    tmp = {}
    for line in device.split("\n"):
        for attrib in _geom_attribs:
            search = re.search(fr"{attrib}:\s(.*)", line)
            if search:
                value = _datavalue(GEOM._datatypes.get(attrib), search.group(1))
                tmp[attrib] = value
                if attrib in GEOM._aliases:
                    tmp[GEOM._aliases[attrib]] = value

    name = tmp.pop(GEOM.GEOMNAME)
    if not name.startswith("cd"):
        disks[name] = tmp
        if not tmp.get(GEOM.ROTATIONRATE):
            log.debug("Device %s reports itself as an SSD", device)
            ssds.append(name)

    return disks, ssds


async def load_disks(hub):
    """
    Return list of disk devices and work out if they are SSD or HDD.
    """
    ssds = []
    disks = {}

    geom = shutil.which("geom")
    if geom:
        devices = (
            (await hub.exec.cmd.run([geom, "disk", "list"]))
            .stdout.strip()
            .split("\n\n")
        )
        for device in devices:
            d, s = await _parse_geom_attribs(device)
            disks.update(d)
            ssds.extend(s)

    if ssds:
        hub.corn.CORN.SSDs = sorted(ssds)
    if disks:
        # This should be a list like it is on every other OS
        # It should also only contain rotational disks like every other OS
        hub.corn.CORN.disks = sorted(set(disks) - set(ssds))



import os
import sys


async def load_cwd(hub):
    """
    Current working directory
    """
    hub.corn.CORN.cwd = os.getcwd()


async def load_path(hub):
    """
    Return the path
    """
    # Provides:
    #   path
    hub.corn.CORN.path = os.environ.get("PATH", "")


async def load_pythonpath(hub):
    """
    Return the Python path
    """
    # Provides:
    #   pythonpath
    hub.corn.CORN.pythonpath = sorted(sys.path, key=str.casefold)


async def load_executable(hub):
    """
    Return the python executable in use
    """
    # Provides:
    #   pythonexecutable
    hub.corn.CORN.pythonexecutable = sys.executable


async def load_shell(hub):
    hub.corn.CORN.shell = os.environ.get("SHELL", "/bin/sh")



import ifcfg


async def load_interfaces(hub):
    """
    Provide a dict of the connected interfaces and their ip addresses
    The addresses will be passed as a list for each interface
    """
    ipv4 = set()
    ipv6 = set()
    ifconfig = ifcfg.parser.UnixParser()

    for interface, device in sorted(ifconfig.interfaces.items()):
        hw_addr = device.get("ether")
        if hw_addr:
            hub.corn.CORN.hwaddr_interfaces[interface] = hw_addr
        ipv4.update(device["inet4"])
        if device["inet4"]:
            hub.corn.CORN.ip4_interfaces[interface] = sorted(device["inet4"])
        ipv6.update(device["inet6"])
        if device["inet6"]:
            hub.corn.CORN.ip6_interfaces[interface] = sorted(device["inet6"])

        hub.corn.CORN.ip_interfaces[interface] = sorted(device["inet4"]) + sorted(
            device["inet6"]
        )

    hub.corn.CORN.ipv4 = sorted(ipv4)
    hub.corn.CORN.ipv6 = sorted(ipv6)



import shutil


async def load_windows_domain(hub):
    if shutil.which("realm"):
        realms = (await hub.exec.cmd.run(["realm", "list", "--name-only"]))[
            "stdout"
        ].splitlines()
        if realms:
            hub.corn.CORN.windowsdomain = realms[0]
            hub.corn.CORN.windowsdomaintype = "Domain"



import logging
import shutil

log = logging.getLogger(__name__)


async def _load_osrelease_freebsd(hub) -> str:
    version = shutil.which("freebsd-version")
    if version:
        return (await hub.exec.cmd.run([version, "-u"])).stdout


async def load_osbuild(hub):
    sysctl = shutil.which("sysctl")

    if sysctl:
        hub.corn.CORN.osbuild = (
            await hub.exec.cmd.run([sysctl, "-n", "kern.osrevision"])
        ).stdout.strip()
    else:
        hub.corn.CORN.osbuild = "unknown"


async def load_oscodename(hub):
    """
    BSD Doesn't have codenames so we'll use the BSD version of the user environment.
    """
    uname = shutil.which("uname")
    if uname:
        hub.corn.CORN.oscodename = (
            await hub.exec.cmd.run([uname, "-U"])
        ).stdout.strip()
    else:
        hub.corn.CORN.oscodename = "unknown"


async def load_osinfo(hub):
    hub.corn.CORN.os = hub.corn.CORN.kernel
    full_release = await _load_osrelease_freebsd(hub) or hub.corn.CORN.kernelrelease
    hub.corn.CORN.osmanufactuer = "unknown"
    hub.corn.CORN.osrelease = full_release.split("-", 1)[0]
    hub.corn.CORN.osfullname = f"{hub.corn.CORN.os}-{full_release}"
    hub.corn.CORN.osrelease_info = tuple(
        int(x) if x.isdigit() else x for x in hub.corn.CORN.osrelease.split(".")
    )
    hub.corn.CORN.osmajorrelease = int(hub.corn.CORN.osrelease_info[0])
    hub.corn.CORN.osfinger = f"{hub.corn.CORN.os}-{hub.corn.CORN.osmajorrelease}"



import logging
import re
import shutil
from typing import Dict, List

log = logging.getLogger(__name__)

KNOWN_VENDORS = [
    "nvidia",
    "amd",
    "ati",
    "intel",
    "cirrus logic",
    "vmware",
    "matrox",
    "aspeed",
]


async def _load_pcictl(hub) -> List[Dict[str, str]]:
    """
    num_gpus: int
    gpus:
      - vendor: nvidia|amd|ati|...
        model: string
    """
    gpus = []

    pcictl = shutil.which("pcictl")
    if pcictl:
        out = await hub.exec.cmd.run([pcictl, "pci0", "list"])
        for line in out.stdout.strip().splitlines():
            for vendor in KNOWN_VENDORS:
                vendor_match = re.match(
                    fr"[0-9:]+ ({vendor}) (.+) \(VGA .+\)", line, re.IGNORECASE
                )
                if vendor_match:
                    gpus.append(
                        {
                            "vendor": vendor_match.group(1),
                            "model": vendor_match.group(2),
                        }
                    )
    return gpus


async def _load_pciconf(hub) -> List[Dict[str, str]]:
    gpus = []
    pciconf = shutil.which("pciconf")
    if pciconf:
        ret = await hub.exec.cmd.run([pciconf, "-l", "-v"])
        if not ret.retcode:
            device = {"model": None, "vendor": "unknown"}
            for line in ret.stdout.splitlines():
                if " = " in line:
                    key, val = line.split("=", 1)
                    key = key.strip()
                    val = val.strip("'\" ")
                    if key == "vendor":
                        device["vendor"] = val
                    elif key == "device":
                        device["model"] = val
                    elif key == "class" and val == "display":
                        gpus.append(device.copy())
    return gpus


async def _load_glxinfo(hub) -> List[Dict[str, str]]:
    """Mesa library for detecting gpu info"""
    gpus = []
    glxinfo = shutil.which("glxinfo")
    if glxinfo:
        ret = await hub.exec.cmd.run([glxinfo, "-B"])
        if not ret.retcode:
            device = {}
            for line in ret.stdout.splitlines():
                if ":" in line:
                    key, val = line.split(":", 1)
                    if key == "OpenGL vendor string":
                        device["vendor"] = val
                    elif key == "OpenGL renderer string":
                        device["model"] = val
            # There will only be one
            if device:
                gpus.append(device)
    return gpus


async def load_gpudata(hub):
    hub.corn.CORN.gpus = (
        await _load_pciconf(hub) or await _load_pcictl(hub) or await _load_glxinfo(hub)
    )
    hub.corn.CORN.num_gpus = len(hub.corn.CORN.gpus)



import aiofiles
import logging
import os
import yaml

log = logging.getLogger(__name__)


async def load_config(hub):
    """
    Return the corn set in the corn file
    """
    # TODO this comes from salt.salt.corn.extra.py, how do we want to locate static corn configs in hub.OPT?
    if "conf_file" not in hub.OPT:
        return

    if os.path.isdir(hub.OPT.conf_file):
        gfn = os.path.join(hub.OPT.conf_file, "corn")
    else:
        gfn = os.path.join(os.path.dirname(hub.OPT.conf_file), "corn")

    if os.path.isfile(gfn):
        log.debug("Loading static corn from %s", gfn)
        async with aiofiles.open(gfn, "rb") as fp_:
            try:
                # TODO? Load the yaml file with hub.rend
                hub.corn.CORN.update(yaml.safe_load(fp_))
            except Exception:  # pylint: disable=broad-except
                log.warning("Bad syntax in corn file! Skipping.")
